%%% Load options file

:- consult('options.pl').

%%%
%%% Auxiliary dynamic predicates
%%%

%%% The following dynamics predicates are used to store input data

:- dynamic	hyponym/2, block/2, isA/2, disjoint/2, mapping/2. %/

% WordNet Hierarchy			->	hyponym/2
%	Usage: hyponyn(S1,S2) if S1 is a hyponym of S1
% WordNet InheritanceBlocks		->	block/2
%	Usage: block(S1,S2) if the inheritance from S1 to S2 is blocked
% TCO Hierarchy				->	isA/2
%	Usage: isA(C1,C2) if C1 is a subclass of C2
% TCO Incompatibilities			->	disjoint/2
%	Usage: disjoint(C1,C2) if the classes C1 and C2 are incompatible
% WordNet to TCO Mapping		->	mapping/2
%	Usage: mapping(S,C) if the synset S is mapped to the class C
%	"mapping/2" relations are not really stored. Instead, we use "mapping/3".

%%% Besides, we use the following dynamic predicates during the execution process:

:- dynamic	mapping/3, inconsistency/5, indexTCO/2. %/ 

% WordNet to TCO Mapping (expanded)	-> 	"mapping/3"
%	Usage: mapping(S,C,L) if the synset S is mapped to the class C using L-1 inheritance steps (in WordNet hierarchy)
% Inconsistency description		-> 	"inconsistency/5"
%	Usage: inconsistency(S,C1,L1,C2,L2) if the synset S is mapped to the class C1/C2 using L1/L2 inheritance steps (in WordNet hierarchy)
% Index of the TCO feature		->	"indexTCO/2"
%	Usage: indexTCO(C,IC) if the feature C has index IC, where IC is of the form 'T.L.O', 'T' is the index of the top feature of C, 'L' is the level of C and 'O' is the order of C among the features of level 'L'

%%%
%%% End of Auxiliary dynamic predicates
%%%

%%%
%%% Main predicates
%%%

% %
% % Consistency checker
% % 

%%% checkAnnotation/0 %/
%%%	Default input and output files are used

checkAnnotation:-
		getWordNetHierarchyFile(F1),
		getWordNetInheritanceBlocksFile(F2),
		getTCOHierarchyFile(F3),
		getTCOIncompatibilitiesFile(F4),
		getWordNet2TCOMappingFile(F5),
		getInconsistenciesFile(NF),
		checkAnnotation(F1,F2,F3,F4,F5,NF).

%%% checkAnnotation/1 %/
%%%	Default input files are used
%%%	checkAnnotation(in NF)
%%%	NF: Inconsistencies file name

checkAnnotation(NF):-
		getWordNetHierarchyFile(F1),
		getWordNetInheritanceBlocksFile(F2),
		getTCOHierarchyFile(F3),
		getTCOIncompatibilitiesFile(F4),
		getWordNet2TCOMappingFile(F5),
		checkAnnotation(F1,F2,F3,F4,F5,NF).

%%% checkAnnotation/6 %/
%%%	checkAnnotation(in F1, in F2, in F3, in F4, in F5, in NF)
%%%	F1: WordNet Hierarchy file name
%%%	F2: WordNet Inheritance Blocks file name
%%%	F3: TCO Hierarchy file name
%%%	F4: TCO Incompatibilities file name
%%%	F5: WordNet to TCO Mapping file name
%%%	NF: Inconsistencies file name

checkAnnotation(F1,F2,F3,F4,F5,NF):-
		removePreviousData,
		readInputData(F1,F2,F3,F4,F5),
		mappingExpansionByTCO,
		lookForInconsistencies,
		writeInconsistencies(NF).

% %
% % WordNet to TCO Annotation
% %

%
% According to "isA/2" and "hyponym/2" relations
%

%%% getTopOntology/0 %/
%%%	Default input and output files are used

getTopOntology:-
		getWordNetHierarchyFile(F1),
		getWordNetInheritanceBlocksFile(F2),
		getTCOHierarchyFile(F3),
		getTCOIncompatibilitiesFile(F4),
		getWordNet2TCOMappingFile(F5),
		getTopOntologyFile(NF),
		getTopOntology(F1,F2,F3,F4,F5,NF).

%%% getTopOntology/1 %/
%%%	Default input files are used
%%%	getTopOntology(in NF)
%%%	NF: Top Ontology file name

getTopOntology(NF):-
		getWordNetHierarchyFile(F1),
		getWordNetInheritanceBlocksFile(F2),
		getTCOHierarchyFile(F3),
		getTCOIncompatibilitiesFile(F4),
		getWordNet2TCOMappingFile(F5),
		getTopOntology(F1,F2,F3,F4,F5,NF).

%%% getTopOntology/6 %/
%%%	getTopOntology(in F1, in F2, in F3, in F4, in F5, in NF)
%%%	F1: WordNet Hierarchy file name
%%%	F2: WordNet Inheritance Blocks file name
%%%	F3: TCO Hierarchy file name
%%%	F4: TCO Incompatibilities file name
%%%	F5: WordNet to TCO Mapping file name
%%%	NF: Top Ontology file name

getTopOntology(F1,F2,F3,F4,F5,NF):-
		removePreviousData,
		readInputData(F1,F2,F3,F4,F5),
		mappingExpansionByTCO,
		lookForInconsistencies,
		writeTopOntology(NF).

%
% Exclusively according to "hyponym/2" relation
%

%%% getExpandedMapping/0 %/
%%%	Default input and output files are used

getExpandedMapping:-
		getWordNetHierarchyFile(F1),
		getWordNetInheritanceBlocksFile(F2),
		getTCOHierarchyFile(F3),
		getTCOIncompatibilitiesFile(F4),
		getWordNet2TCOMappingFile(F5),
		getExpandedMappingFile(NF),
		getExpandedMapping(F1,F2,F3,F4,F5,NF).

%%% getExpandedMapping/1 %/
%%%	Default input files are used
%%%	getExpandedMapping(in NF)
%%%	NF: Expanded mapping file name

getExpandedMapping(NF):-
		getWordNetHierarchyFile(F1),
		getWordNetInheritanceBlocksFile(F2),
		getTCOHierarchyFile(F3),
		getTCOIncompatibilitiesFile(F4),
		getWordNet2TCOMappingFile(F5),
		getExpandedMapping(F1,F2,F3,F4,F5,NF).

%%% getExpandedMapping/6 %/
%%%	getExpandedMapping(in F1, in F2, in F3, in F4, in F5, in NF)
%%%	F1: WordNet Hierarchy file name
%%%	F2: WordNet Inheritance Blocks file name
%%%	F3: TCO Hierarchy file name
%%%	F4: TCO Incompatibilities file name
%%%	F5: WordNet to TCO Mapping file name
%%%	NF: Expanded mapping file name

getExpandedMapping(F1,F2,F3,F4,F5,NF):-
		removePreviousData,
		readInputData(F1,F2,F3,F4,F5),
		lookForInconsistencies,
		writeExpandedMapping(NF).

%%%
%%% End of Main predicates
%%%

%%% 
%%% Auxiliary predicates
%%%

% %
% % Remove previous data
% %

%%% removePreviousData/0 %/
%%%	Output: remove data from previous executions

removePreviousData:-
		retractall((hyponym(_,_))),
		retractall((block(_,_))),
		retractall((isA(_,_))),
		retractall((disjoint(_,_))),
		retractall((mapping(_,_,_))),
		retractall((inconsistency(_,_,_,_,_))),
		retractall((indexTCO(_,_))).

% %
% % Read input data
% %

%%% readInputData(in F1, in F2, in F3, in F4, in F5)
%%%	F1: WordNet Hierarchy file name
%%%	F2: WordNet Inheritance Blocks file name
%%%	F3: TCO Hierarchy file name
%%%	F4: TCO Incompatibilities file name
%%%	F5: WordNet to TCO Mapping file name
%%%	Output: input data is stored using dynamic predicates

readInputData(F1,F2,F3,F4,F5):-
		readWordNetHierarchy(F1),
		readWordNetInheritanceBlocks(F2),
		readTCOHierarchy(F3),
		readTCOIncompatibilities(F4),
		readWordNet2TCOMapping(F5).

%%% readWordNetHierarchy(in F)
%%%	F: WordNet Hierarchy file name
%%%	Output: input WordNet Hierarchy data is stored using "hyponym/2"

readWordNetHierarchy(F):-
		catch(open(F,read,S),_,inputDataError(F,'WordNet Hierarchy')),
		read(S,L),
		parseWordNetHierarchy(S,L),
		close(S).

%%% parseWordNetHierarchy(in S, in L)
%%%	S: WordNet Hierarchy stream
%%%	L: next line
%%%	Output: input WordNet Hierarchy data is stored using "hyponym/2"

parseWordNetHierarchy(_,end_of_file):-
		!.
parseWordNetHierarchy(S,hyponym(S1,S2)):-
		!,
		assert((hyponym(S1,S2))),
		read(S,L),
		parseWordNetHierarchy(S,L).
parseWordNetHierarchy(_,L):-
		print('\n'),
		print('Incorrect information in WordNet Hierarchy file --> '),
		print(L),
		fail.

%%% readWordNetInheritanceBlocks(in F)
%%%	F: WordNet Inheritance Blocks file name
%%%	Output: input WordNet Inheritance Blocks data is used to remove the corresponding "hyponym/2" relations

readWordNetInheritanceBlocks(F):-
		catch(open(F,read,S),_,inputDataError(F,'WordNet Inheritance Blocks')),
		read(S,L),
		parseWordNetInheritanceBlocks(S,L),
		close(S).

%%% parseWordNetInheritanceBlocks(in S, in L)
%%%	S: WordNet Inheritance Blocks stream
%%%	L: next line
%%%	Output: input WordNet Inheritance Blocks data is used to remove the corresponding "hyponym/2" relations

parseWordNetInheritanceBlocks(_,end_of_file):-
		!.
parseWordNetInheritanceBlocks(S,block(S1,S2)):-
		retract((hyponym(S2,S1))),
		!,
		read(S,L),
		parseWordNetInheritanceBlocks(S,L).
parseWordNetInheritanceBlocks(S,block(S1,S2)):-
		print('\n'),
		print('Warning --> '),
		print(block(S1,S2)),
		print(' inheritance block point is never used'),
		print('\n'),
		read(S,L),
		parseWordNetInheritanceBlocks(S,L).
parseWordNetInheritanceBlocks(_,L):-
		print('\n'),
		print('Incorrect information in WordNet Inheritance Blocks file --> '),
		print(L),
		fail.

%%% readTCOHierarchy(in F)
%%%	F: TCO Hierarchy file name
%%%	Output: input TCO Hierarchy data is stored using "isA/2"

readTCOHierarchy(F):-
		catch(open(F,read,S),_,inputDataError(F,'TCO Hierarchy')),
		read(S,L),
		parseTCOHierarchy(S,L),
		close(S).

%%% parseTCOHierarchy(in S, in L)
%%%	S: TCO Hierarchy stream
%%%	L: next line
%%%	Output: input TCO Hierarchy data is stored using "isA/2"

parseTCOHierarchy(_,end_of_file):-
		!.
parseTCOHierarchy(S,isA(C1,C2)):-
		!,
		assert((isA(C1,C2))),
		read(S,L),
		parseTCOHierarchy(S,L).
parseTCOHierarchy(_,L):-
		print('\n'),
		print('Incorrect information in TCO Hierarchy file --> '),
		print(L),
		fail.

%%% readTCOIncompatibilities(in F)
%%%	F: TCO Incompatibilities file name
%%%	Output: input TCO Incompatibilities data is stored using "disjoint/2"

readTCOIncompatibilities(F):-
		catch(open(F,read,S),_,inputDataError(F,'TCO Incompatibilities')),
		read(S,L),
		parseTCOIncompatibilities(S,L),
		close(S).

%%% parseTCOIncompatibilities(in S, in L)
%%%	S: TCO Incompatibilities stream
%%%	L: next line
%%%	Output: input TCO Incompatibilities data is stored using "disjoint/2"

parseTCOIncompatibilities(_,end_of_file):-
		!.
parseTCOIncompatibilities(S,disjoint(C1,C2)):-
		!,
		assert((disjoint(C1,C2))),
		read(S,L),
		parseTCOIncompatibilities(S,L).
parseTCOIncompatibilities(_,L):-
		print('\n'),
		print('Incorrect information in TCO Incompatibilities file --> '),
		print(L),
		fail.

%%% readWordNet2TCOMapping(in F)
%%%	F: WordNet to TCO Mapping file name
%%%	Output: input WordNet to TCO Mapping data is stored using "mapping/3"

readWordNet2TCOMapping(F):-
		catch(open(F,read,S),_,inputDataError(F,'WordNet to TCO Mapping')),
		read(S,L),
		parseWordNet2TCOMapping(S,L),
		close(S).

%%% parseWordNet2TCOMapping(in S, in L)
%%%	S: WordNet 2 TCO Mapping stream
%%%	L: next line
%%%	Output: input WordNet to TCO Mapping data is stored using "mapping/3"

parseWordNet2TCOMapping(_,end_of_file):-
		!.
parseWordNet2TCOMapping(SF,mapping(S,C)):-
		(
		 (
		  isA(C,_)
		 );
		 (
		  isA(_,C)
		 )
		),
		!,
		assert((mapping(S,C,1))),
		read(SF,L),
		parseWordNet2TCOMapping(SF,L).
parseWordNet2TCOMapping(_,mapping(S,C)):-
		!,
		print('\n'),
		print('Incorrect information in WordNet to TCO Mapping file --> the concept '),
		print(C),
		print(' in the mapping '),
		print(mapping(S,C)),
		print(' does not exist.'),
		fail.
parseWordNet2TCOMapping(_,L):-
		print('\n'),
		print('Incorrect information in WordNet to TCO Mapping file --> '),
		print(L),
		fail.

%%% inputDataError(in F, in D)
%%%	F: the file that we have tried to open
%%%	D: the data that we have tried to read
%%%	Action: an error has occurred while opening the file. An error message is printed and the program fails

inputDataError(F,D):-
		print('\n'),
		print('The file '),
		print(F),
		print(' does not exist -> '),
		print(D),
		print(' file name'),
		fail.

% %
% % Mapping Expansion By TCO
% %

%%% mappingExpansionByTCO
%%%	Output: "mapping/3" relations that are obtained by inheritance of the relation "isA/2"

mappingExpansionByTCO:-
		findall(mapping(S,C,1),transitivityIsA(S,C),L),
		parseMappingExpansionByTCO(L).

%%% transitivityIsA(out S, out C)
%%%	S: a synset of WordNet
%%%	C: a class of TCO
%%%	Action: S is mapped to C according to the transitivity of the relation "isA/2"
%%%		S was not previously mapped to C

transitivityIsA(S,C):-
		mapping(S,NC,1),
		isA(NC,C),
		not(mapping(S,C,1)).

%%% parseMappingExpansionByTCO(in L)
%%%	L: the list of new mappings of WordNet to TCO that are obtained by inheritance of the relation "isA/2"
%%%	Action:	if L is the empty list -> 		STOP
%%%		if L is NOT the empty list ->		the mappings in L are added
%%%							we look for new mappings

parseMappingExpansionByTCO([]):-
		!.
parseMappingExpansionByTCO(L):-
		addNewMappings(L),
		findall(mapping(S,C,1),transitivityIsA(S,C),NL),
		parseMappingExpansionByTCO(NL).

%%% addNewMappings(in L)
%%%	L: the list of new mappings of WordNet to TCO that are obtained by inheritance of the relations "isA/2" and "hyponym/2"
%%%	Action: the mappings in L are added

addNewMappings([]).
addNewMappings([mapping(S,C,L)|RL]):-
		mapping(S,C,L),			% For eliminating the repeated elements returned by a call to "findall/3"
		!,
		addNewMappings(RL).
addNewMappings([mapping(S,C,L)|RL]):-
		assert((mapping(S,C,L))),
		addNewMappings(RL).

% %
% % Look for inconsistencies
% % 

%%% lookForInconsistencies
%%%	Action: we look for synsets that are mapped to non-compatible TCO classes
%%%		The search is organized by levels in "hyponym/2" inheritance for efficiency purposes

lookForInconsistencies:-
		lookForInconsistenciesByLevels(1).

%%% lookForInconsistenciesByLevels(in L)
%%%	L: the current level of "hyponym/2" inheritance
%%%	Action:	1. We look for inconsistencies of level L
%%%		2. We expand the mapping to level L+1.
%%%			If the expansion is empty, STOP
%%%			If the expansion is NOT empty, we look for inconsistencies of level L+1

lookForInconsistenciesByLevels(L):-
		findall(inconsistency(S,C1,L1,C2,L2),inconsistencyOfCurrentLevel(L,S,C1,L1,C2,L2),LI),
		parseInconsistencies(LI),
		NL is L+1,
		mappingExpansionByWordNet(NL,LM),
		(
		 (
		  LM = []
		 );
		 (
		  LM \= [],
		  lookForInconsistenciesByLevels(NL)
		 )
		).

%%% inconsistencyOfCurrentLevel(in L, out S, out C1, out L1, out C2, out L2)
%%%	L: current level in "hyponym/2" inheritance
%%%	S: synset that is mapped to non-compatible classes C1 and C2
%%%	C1: a TCO class
%%%	L1: level in "hyponym/2" inheritance for mapping S to C1
%%%	C2: a tCO class
%%%	L2: level in "hyponym/2" inheritance for mapping S to C2

inconsistencyOfCurrentLevel(L1,S,C1,L1,C2,L2):-
		mapping(S,C1,L1),
		mapping(S,C2,L2),
		disjoint(C1,C2).
inconsistencyOfCurrentLevel(L1,S,C1,L1,C2,L2):-
		mapping(S,C1,L1),
		mapping(S,C2,L2),
		disjoint(C2,C1),
		L1 > L2.

%%% parseInconsistencies(in L)
%%%	L: a list of inconsistencies
%%%	Action: the inconsistencies in L are stored using the relation "inconsistency/5"

parseInconsistencies([]).
parseInconsistencies([inconsistency(S,C1,L1,C2,L2)|L]):-
		inconsistency(S,C1,L1,C2,L2),			% For eliminating the repeated elements returned by a call to "findall/3"
		!,
		parseInconsistencies(L).
parseInconsistencies([inconsistency(S,C1,L1,C2,L2)|L]):-
		assert((inconsistency(S,C1,L1,C2,L2))),
		parseInconsistencies(L).

%%% mappingExpansionByWordNet(in L, out LM)
%%%	L: level in "hyponym/2" inheritance
%%%	LM: list of mappings of level L in "hyponym/2" inheritance
%%%	Action: the mapping chains that currently yield to an inconsistency are discarded

mappingExpansionByWordNet(L,LM):-
		NL is L-1,
		findall(mapping(S,C,L),transitivityHyponym(NL,S,C),LM),
		addNewMappings(LM).

%%% transitivityHyponym(in L, out S, out C)
%%%	L: level in "hyponym/2" inheritance
%%%	S: a synset of WordNet
%%%	C: a class of TCO
%%%	Action: S is mapped to C according to the transitivity of the relation "hyponym/2" in level L+1
%%%		S was not previously mapped to C

transitivityHyponym(L,S,C):-
		hyponym(S,NS),
		mapping(NS,C,L),
    		not(mapping(S,C,_)),
		not(inconsistency(NS,C,_,_,_)),
		not(inconsistency(NS,_,_,C,_)).

% %
% % Write inconsistencies
% %

%%% writeInconsistencies(in F)
%%%	F: Inconsistencies file name

writeInconsistencies(F):-
		open(F,write,SF),
		findall(inconsistency(S,C1,L1,C2,L2),inconsistency(S,C1,L1,C2,L2),L),
		writeListOfInconsistencies(SF,L),
		length(L,N),
		print(N),
		print(' inconsistencies found.'),
		close(SF).

%%% writeListOfInconsistencies(in S, in L)
%%%	S: inconsistencies stream
%%%	L: list of inconsistencies
%%%	Output: the inconsistencies in L are written in S

writeListOfInconsistencies(_,[]).
writeListOfInconsistencies(SF,[inconsistency(S,C1,L1,C2,L2)|L]):-
		print(SF,inconsistency(S,C1,C2)),
		print(SF,'\n'),
		print(SF,'\t'),
		print(SF,C1),
		print(SF,'\n'),
		writeHyponymPath(SF,S,C1,L1),
		print(SF,'\t'),
		print(SF,C2),
		print(SF,'\n'),
		writeHyponymPath(SF,S,C2,L2),
		print(SF,'\n'),
		writeListOfInconsistencies(SF,L).

%%% writeHyponymPath(in SF, in S, in C, in L)
%%%	SF: inconsistencies stream
%%%	S: a synset
%%%	C: a class of TCO
%%%	L: level in "hyponym/2" inheritance

writeHyponymPath(SF,S,C,L):-
		findHyponymPath(S,C,L,P),
		reverse(P,NP),
		printHyponymPath(SF,NP).

%%% findHyponymPath(in S, in C, in L, out P)
%%%	S: a synset
%%%	C: a class of TCO
%%%	L: level in "hyponym/2" inheritance

findHyponymPath(S,C,1,[S]):-
		mapping(S,C,1),
		!.
findHyponymPath(S,C,N,[S|P]):-
		N > 1,
		hyponym(S,NS),
		NN is N-1,
		findHyponymPath(NS,C,NN,P).

%%% printHyponymPath(in S, in P)
%%%	S: inconsistencies stream
%%%	P: a path of "hyponym/2" inheritance

printHyponymPath(_,[]).
printHyponymPath(SF,[S|P]):-
		print(SF,'\t'),
		print(SF,'\t'),
		print(SF,S),
		print(SF,'\t'),
		length([S|P],N),
		print(SF,N),
		print(SF,'\n'),
		printHyponymPath(SF,P).

% %
% % Write Top Ontolgy
% %

%%% writeTopOntology(in F)
%%%	F: Top Ontology file name

writeTopOntology(F):-
		findall(inconsistency(S,C1,L1,C2,L2),inconsistency(S,C1,L1,C2,L2),LI),
		(
		 (
		  LI = [],
		  !,
		  open(F,write,SF),
		  print(SF,'This resource is free software: you can redistribute it and/or modify\n'),
		  print(SF,'   it under the terms of the GNU General Public License as published by\n'),
		  print(SF,'   the Free Software Foundation, either version 3 of the License, or\n'),
		  print(SF,'   (at your option) any later version.\n'),
		  print(SF,'\n'),
		  print(SF,'   This resource is distributed in the hope that it will be useful,\n'),
		  print(SF,'   but WITHOUT ANY WARRANTY; without even the implied warranty of\n'),
		  print(SF,'   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n'),
		  print(SF,'   GNU General Public License for more details.\n'),
		  print(SF,'\n'),
		  print(SF,'   You should have received a copy of the GNU General Public License\n'),
		  print(SF,'   along with this resource.  If not, see <http://www.gnu.org/licenses/>.\n'),
		  print(SF,'\n'),
		  print(SF,'Copyright 2008 Javier lvez, Jordi Atserias, Jordi Carrera, Salvador Climent,\n'),
		  print(SF,'   Egoitz Laparra, Joaquim Mor, Antoni Oliver and German Rigau\n'),
		  print(SF,'\n'),
		  print(SF,'==============================================================================================\n'),
		  getIndexTCO,
%		  findall(S,mapping(S,_,_),LS),
%		  remove_duplicates(LS,NLS),
%		  sort(NLS,NLSO),
		  findall((S,IC),(mapping(S,C,_),indexTCO(C,IC)),LM),
		  sort(LM,NLM),
		  writeListOfMappingsTopOntology(SF,NLM),
		  close(SF)
		 );
		 (
		  print('There exists some inconsistencies in the mapping from WordNet to TCO.'),
		  print('\n'),
		  print('Please correct it before using this command.')
		 )
		).

%%% writeListOfMappingsTopOntology(in S, in L) 
%%%	S: Top Ontology stream
%%%	L: list of mappings

writeListOfMappingsTopOntology(_,[]).
writeListOfMappingsTopOntology(SF,[(S,IC)|L]):-
		print(SF,'\n'),
		print(SF,S),
		print(SF,'|'),
		indexTCO(C,IC),
		!,					% This cut symbol is strictly necessary to save local stack, but it has no side-effects
		print(SF,C),
		writeListOfMappingsTopOntology(SF,L).

%%% getIndexTCO/0 %/

getIndexTCO:-
		findall(C,(isA(C,_);isA(_,C)),LC),
		remove_duplicates(LC,NLC),
		getIndexTCOList(NLC).

%%% remove_duplicates(in L, out NL)
%%%	L: a list
%%%	NL: the list that results from removing the duplicated elements from L

remove_duplicates([],[]).
remove_duplicates([E|L1],[E|L2]):-
		delete(L1,E,NL1),
		remove_duplicates(NL1,L2).

%%% getIndexTCOList(in LF)
%%%	LF: a list of TCO features

getIndexTCOList([]).
getIndexTCOList([C|LF]):-
		getTCOIndexFeature(C,LI),
		reverse(LI,LC),
		flatten(LC,NLC),
		atom_chars(IC,NLC),
		assert((indexTCO(C,IC))),
		getIndexTCOList(LF).

%%% getTCOIndexFeature(in C, out LI)
%%%	C: a TCO feature
%%%	LI: the list of index of C and its ancestors

getTCOIndexFeature(C,[I,'.'|LI]):-
		isA(C,NC),
		!,
		getFeaturesOfSameLevel(C,LF),
		nth0(T,LF,C),
		number_chars(T,LC),
		twoDigits(LC,I),
		getTCOIndexFeature(NC,LI).
getTCOIndexFeature(C,[I]):-
		getTCOTopFeatures(LF),
		nth0(T,LF,C),
		number_chars(T,LC),
		twoDigits(LC,I).

%%% getFeaturesOfSameLevel(in C, out LF)
%%%	C: a TCO feature
%%%	LF: a list of features of the same level than C

getFeaturesOfSameLevel(C,NLF):-
		isA(C,NC),
		findall(F,isA(F,NC),LF),
		LF \= [],
		!,
		sort(LF,NLF).
getFeaturesOfSameLevel(C,[C]).

%%% getTCOTopFeatures(out LF)
%%%	LF: the list of features in TCO without ancestor

getTCOTopFeatures(NLF):-
		findall(C,(isA(_,C), not(isA(C,_))),LF),
		sort(LF,NLF).

%%% twoDigits(in LC, out NLC)
%%%	LC: a list of characters that represents a number
%%%	NLC: the list of characters that represents the same number than LC using two digits
%%%	Note: it fails for numbers that need more than two characters

twoDigits([C],['0',C]):-
		!.
twoDigits([C1,C2],[C1,C2]):-
		!.
twoDigits(_,_):-
		print('Two many TCO features for two digits.'),
		fail.

% %
% % Write expanded mapping
% %

%%% writeExpandedMapping(in F)
%%%	F: expanded mapping file name

writeExpandedMapping(F):-
		findall(inconsistency(S,C1,L1,C2,L2),inconsistency(S,C1,L1,C2,L2),LI),
		(
		 (
		  LI = [],
		  !,
		  open(F,write,SF),
		  print(SF,'This resource is free software: you can redistribute it and/or modify\n'),
		  print(SF,'   it under the terms of the GNU General Public License as published by\n'),
		  print(SF,'   the Free Software Foundation, either version 3 of the License, or\n'),
		  print(SF,'   (at your option) any later version.\n'),
		  print(SF,'\n'),
		  print(SF,'   This resource is distributed in the hope that it will be useful,\n'),
		  print(SF,'   but WITHOUT ANY WARRANTY; without even the implied warranty of\n'),
		  print(SF,'   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n'),
		  print(SF,'   GNU General Public License for more details.\n'),
		  print(SF,'\n'),
		  print(SF,'   You should have received a copy of the GNU General Public License\n'),
		  print(SF,'   along with this resource.  If not, see <http://www.gnu.org/licenses/>.\n'),
		  print(SF,'\n'),
		  print(SF,'Copyright 2008 Javier lvez, Jordi Atserias, Jordi Carrera, Salvador Climent,\n'),
		  print(SF,'   Egoitz Laparra, Joaquim Mor, Antoni Oliver and German Rigau\n'),
		  print(SF,'\n'),
		  print(SF,'==============================================================================================\n'),
		  findall((C,S,N),mapping(S,C,N),LM),
		  splitListOfMappings(LM,LM1,LM2),
		  sort(LM1,NLM1),
		  writeListOfMappings(SF,NLM1,'='),
		  sort(LM2,NLM2),
		  writeListOfMappings(SF,NLM2,'+'),
		  close(SF)
		 );
		 (
		  print('There exists some inconsistencies in the mapping from WordNet to TCO.'),
		  print('\n'),
		  print('Please correct it before using this command.')
		 )
		).

%%% splitListOfMappings(in LM, out LM1, out LM2)
%%%	LM: a list of mappings of the form (C,S,N)
%%%	LM1: the list of direct mappings (N = 1) of the form (C,S) in LM
%%%	LM2: the list of non-direct mappings (N \= 1) of the form (C,S) in LM

splitListOfMappings([],[],[]).
splitListOfMappings([(C,S,1)|LM],[(C,S)|LM1],LM2):-
		!,
		splitListOfMappings(LM,LM1,LM2).
splitListOfMappings([(C,S,_)|LM],LM1,[(C,S)|LM2]):-
		splitListOfMappings(LM,LM1,LM2).

%%% writeLisOfMappings(in S, in L, in D)
%%%	S: Top Ontology stream
%%%	L: a list of mappings
%%%	D: '=' or '+'

writeListOfMappings(_,[],_).
writeListOfMappings(SF,[(C,S)|L],D):-
		print(SF,'\n'),
		print(SF,C),
		print(SF,'|'),
		print(SF,S),
		print(SF,'|n|en16|'),
		print(SF,D),
		writeListOfMappings(SF,L,D).